--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Trans = require("dogma.trans.js.Trans")
local Parser = require("dogma.syn.Parser")

--Suite.
return suite("dogma.trans.js._.StmtTrans", function()
  local trans, parser

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
    trans = Trans.new()
    trans:transform(parser)
  end):title("Create transformer")

  ---------
  -- use --
  ---------
  suite("use", function()
    test("use LiteralStr", function()
      parser:parse('use "my/module"')
      assert(trans:next()):eq('const module = dogma.use(require("my/module"));\n')
    end)

    test("use LiteralStr as Name", function()
      parser:parse('use "my/module" as mod')
      assert(trans:next()):eq('const mod = dogma.use(require("my/module"));\n')
    end)

    test("use LiteralStr, LiteralStr", function()
      parser:parse('use "my/mod1", "my/mod2"')
      assert(trans:next()):eq('const mod1 = dogma.use(require("my/mod1"));const mod2 = dogma.use(require("my/mod2"));\n')
    end)

    test("use LiteralStr as Name, LiteralStr as Name", function()
      parser:parse('use "my/module1" as mod1, "my/module2" as mod2')
      assert(trans:next()):eq('const mod1 = dogma.use(require("my/module1"));const mod2 = dogma.use(require("my/module2"));\n')
    end)

    test("use {variables} = module", function()
      parser:parse('use {var1, var2} = "my/mod1", {var3, var4} = pkg:my/mod2')
      assert(trans:next()):eq('const {var1,var2} = dogma.use(require("my/mod1"));const {var3,var4} = dogma.use(require("my/mod2"));\n')
    end)

    test("use alias:name", function()
      parser:parse("use alias:fs", nil, {imports={fs="@dogmalang/fs2"}})
      assert(trans:next()):eq('const fs = dogma.use(require("@dogmalang/fs2"));\n')
    end)

    test("use fs1 = alias:name as other", function()
      parser:parse("use fs1 = alias:fs", nil, {imports={fs="@dogmalang/fs"}})
      assert(trans:next()):eq('const fs1 = dogma.use(require("@dogmalang/fs"));\n')
    end)
  end):tags("use")

  ----------
  -- from --
  ----------
  suite("from", function()
    test("from LiteralStr use Name", function()
      parser:parse([[from "fs" use FSWatcher]])
      assert(trans:next()):eq('import {FSWatcher} from "fs";\n')
    end)

    test("from LiteralStr use Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher]])
      assert(trans:next()):eq('import {FSWatcher as Watcher} from "fs";\n')
    end)

    test("from LiteralStr use Name, Name", function()
      parser:parse([[from "fs" use FSWatcher, watch]])
      assert(trans:next()):eq('import {FSWatcher, watch} from "fs";\n')
    end)

    test("from LiteralStr use Name as Name, Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher, watch as w]])
      assert(trans:next()):eq('import {FSWatcher as Watcher, watch as w} from "fs";\n')
    end)

    test("from module use Name=>{variables}", function()
      parser:parse([[from "module" use mem1 as MEM1=>{var1,var2}, mem2=>{var3,var4}]])
      assert(trans:next()):eq('import {mem1 as MEM1, mem2} from "module";let {var1,var2} = MEM1;let {var3,var4} = mem2;\n')
    end)
  end)

  -----------
  -- break --
  -----------
  suite("break", function()
    test("break", function()
      parser:parse("break")
      assert(trans:next()):eq("break;\n")
    end)

    test("break for", function(params)
      parser:parse("break " .. params[1])
      assert(trans:next()):eq("$__break__ = true; break;\n")
    end):iter(
      {subtitle=" # for", params="for"},
      {subtitle=" # for each", params="for each"},
      {subtitle=" # while", params="while"},
      {subtitle=" # until", params="until"}
    )
  end)

  ----------
  -- next --
  ----------
  test("next", function()
    parser:parse("next")
    assert(trans:next()):eq("continue;\n")
  end)

  ----------
  -- enum --
  ----------
  suite("enum", function()
    test("export enum Name {Name}", function()
      parser:parse("export enum Color {RED}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
module.exports = exports = Color;Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});

]])
    end)

    test("pub enum Name {Name}", function()
      parser:parse("pub enum Color {RED}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
export {Color};Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});

]])
    end)

    test("enum Name {Name, Name, Name}", function()
      parser:parse("enum Color {RED, GREEN, BLUE}")
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
Object.defineProperty(Color, "RED", {value: new Color("RED", 1), enum: true});
Object.defineProperty(Color, "GREEN", {value: new Color("GREEN", 2), enum: true});
Object.defineProperty(Color, "BLUE", {value: new Color("BLUE", 3), enum: true});

]])
    end)

    test([[enum Name {Name = Str, Name = LiteralStr, Name = LiteralStr}]], function()
      parser:parse([[enum Color {RED = "1", GREEN = "2", BLUE = "3"}]])
      assert(trans:next()):eq([[
class Color {
  constructor(name, val) {
    Object.defineProperty(this, "name", {value: name, enum: true});
    Object.defineProperty(this, "value", {value: val, enum: true});
  }

  /* istanbul ignore next */
  toString() {
    return this.name;
  }
}
Object.defineProperty(Color, "RED", {value: new Color("RED", "1"), enum: true});
Object.defineProperty(Color, "GREEN", {value: new Color("GREEN", "2"), enum: true});
Object.defineProperty(Color, "BLUE", {value: new Color("BLUE", "3"), enum: true});

]])
    end)
  end):tags("enum")

  -----------
  -- async --
  -----------
  suite("async", function()
    test("async without catch", function()
      parser:parse("async\n 1+2\n 3+4")
      assert(trans:next()):eq("setImmediate(async () => {(1+2);(3+4);});\n")
    end)

    test("async with catch without var", function()
      parser:parse("async\n 1+2\n 3+4\ncatch\n 5+6")
      assert(trans:next()):eq("setImmediate(async () => {try {(1+2);(3+4);} catch(_) {(5+6);}});\n")
    end)

    test("async with catch with var", function()
      parser:parse("async\n 1+2\n 3+4\ncatch e\n 5+6")
      assert(trans:next()):eq("setImmediate(async () => {try {(1+2);(3+4);} catch(e) {(5+6);}});\n")
    end)

    test("async with {delay=exp}", function()
      parser:parse("async with {delay = 123*456} 789")
      assert(trans:next()):eq("dogma.asyncly(async () => {789;}, (123*456));\n")
    end)

    test("async with {delay}", function()
      parser:parse("async with {delay} 789")
      assert(trans:next()):eq("dogma.asyncly(async () => {789;}, delay);\n")
    end)

    test("async with {repeat=exp}", function()
      parser:parse("async with {repeat=true} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: null, repeat: true});\n")
    end)

    test("async with {repeat}", function()
      parser:parse("async with {repeat} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: null, repeat: repeat});\n")
    end)

    test("async with {delay=exp, repeat=exp}", function()
      parser:parse("async with {delay=10, repeat=5} 789")
      assert(trans:next()):eq("dogma.repeatAsyncly(async () => {789;}, {delay: 10, repeat: 5});\n")
    end)
  end):tags("async")

  ---------
  -- var --
  ---------
  suite("var", function()
    suite("std", function()
      test("export var Name = Exp", function()
        parser:parse("export var x = 3")
        assert(trans:next()):eq("var x = 3;module.exports = exports = x;\n")
      end)

      test("pub var Name = Exp", function()
        parser:parse("pub var x = 3")
        assert(trans:next()):eq("var x = 3;export {x};\n")
      end)

      test("var Name", function()
        parser:parse("var x")
        assert(trans:next()):eq("let x;\n")
      end)

      test("var Name, Name", function()
        parser:parse("var x, y")
        assert(trans:next()):eq("let x;let y;\n")
      end)

      test("var Name = Exp", function()
        parser:parse("var x = 1+2")
        assert(trans:next()):eq("let x = (1+2);\n")
      end)

      test("var Name = Exp, Name = Exp", function()
        parser:parse("var x = 1+2, y = 3+4")
        assert(trans:next()):eq("let x = (1+2);let y = (3+4);\n")
      end)

      test("var Name = Exp, Name ::= Exp", function()
        parser:parse("var x = 1+2, y ::= 3+4")
        assert(trans:next()):eq("let x = (1+2);const y = (3+4);\n")
      end)
    end)

    suite("map", function()
      test("var {Name} = Exp", function()
        parser:parse("var {x} = 1+2")
        assert(trans:next()):eq("let {x} = (1+2);\n")
      end)

      test("pub var {Name} = Exp", function()
        parser:parse("pub var {x} = 1+2")
        assert(trans:next()):eq("var {x} = (1+2);export {x};\n")
      end)

      test("var {Name, Name} = Exp", function()
        parser:parse("var {x, y} = 1+2")
        assert(trans:next()):eq("let {x, y} = (1+2);\n")
      end)

      test("pub var {Name, Name} = Exp", function()
        parser:parse("pub var {x, y} = 1+2")
        assert(trans:next()):eq("var {x, y} = (1+2);export {x, y};\n")
      end)
    end)

    suite("list", function()
      test("var [Name] = Exp", function()
        parser:parse("var [x] = 1+2")
        assert(trans:next()):eq("let [x] = dogma.getArrayToUnpack((1+2), 1);\n")
      end)

      test("var [Name] ::= Exp", function()
        parser:parse("var [x] ::= 1+2")
        assert(trans:next()):eq("const [x] = dogma.getArrayToUnpack((1+2), 1);\n")
      end)

      test("var [Name, Name] = Exp", function()
        parser:parse("var [x, y] = 1+2")
        assert(trans:next()):eq("let [x, y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("var [_, Name] = Exp", function()
        parser:parse("var [_, y] = 1+2")
        assert(trans:next()):eq("let [, y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("var [Name, _, Name] = Exp", function()
        parser:parse("var [x, _,z] = 1+2")
        assert(trans:next()):eq("let [x, , z] = dogma.getArrayToUnpack((1+2), 3);\n")
      end)

      test("pub var [Name] = Exp", function()
        parser:parse("pub var [x] = 1+2")
        assert(trans:next()):eq("var [x] = dogma.getArrayToUnpack((1+2), 1);export {x};\n")
      end)

      test("var [Name, ...Name] = Exp", function()
        parser:parse("var [x, ...y] = 1+2")
        assert(trans:next()):eq("let [x, ...y] = dogma.getArrayToUnpack((1+2), 2);\n")
      end)

      test("pub var [Name, ...Name] = Exp", function()
        parser:parse("pub var [x, ...y] = 1+2")
        assert(trans:next()):eq("var [x, ...y] = dogma.getArrayToUnpack((1+2), 2);export {x, y};\n")
      end)
    end)
  end):tags("var")

  -----------
  -- const --
  -----------
  suite("const", function()
    test("const Name = Exp", function()
      parser:parse("const x = 1+2")
      assert(trans:next()):eq("const x = (1+2);\n")
    end)

    test("const Name = Exp, Name = Exp", function()
      parser:parse("const x = 1+2, y = 3+4")
      assert(trans:next()):eq("const x = (1+2);const y = (3+4);\n")
    end)

    test("export const Name = Exp", function()
      parser:parse("export const x = 3")
      assert(trans:next()):eq("const x = 3;module.exports = exports = x;\n")
    end)

    test("pub const Name = Exp", function()
      parser:parse("pub const x = 3")
      assert(trans:next()):eq("const x = 3;export {x};\n")
    end)

    test("const {Name} = Exp", function()
      parser:parse("const {x} = 1+2")
      assert(trans:next()):eq("const {x} = (1+2);\n")
    end)

    test("const [Name] = Exp", function()
      parser:parse("const [x] = 1+2")
      assert(trans:next()):eq("const [x] = dogma.getArrayToUnpack((1+2), 1);\n")
    end)
  end):tags("var")

  -----------
  -- while --
  -----------
  suite("while", function()
    test("while Exp do Exp", function()
      parser:parse("while true do x+1")
      assert(trans:next()):eq("while (true) {(x+1);}\n")
    end)

    test("while Exp do Stmt", function()
      parser:parse("while true do return x+1")
      assert(trans:next()):eq("while (true) {return (x+1);}\n")
    end)

    test("while Exp do Block else Block", function()
      parser:parse("while true do x+1\nelse x+2")
      assert(trans:next()):eq("{let $__break__; while (true) {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("while Decl; Exp do Sent", function()
      parser:parse("while x ::= 1; x < 10 do x + 2")
      assert(trans:next()):eq("for (const x=1; (x<10); ) {(x+2);}\n")
    end)

    test("while Decl; Exp do Block else Block", function()
      parser:parse("while x := 1; x < 10 do x + 2\nelse x + 3")
      assert(trans:next()):eq("{let $__break__; for (let x=1; (x<10); ) {(x+2);} if (!$__break__) {(x+3);}}\n")
    end)

    test("while with catch", function()
      parser:parse("while true do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("while (true) try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("while Exp do Block else Block catch Block", function()
      parser:parse("while true do x+1\nelse x+2\ncatch\n x+3")
      assert(trans:next()):eq("try {let $__break__; while (true) {(x+1);} if (!$__break__) {(x+2);}} catch(_) {(x+3);}\n")
    end)

    test("while with finally", function()
      parser:parse("while true do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("while (true) try {(x+1);} finally {(x+2);}\n")
    end)

    test("while with catch and finally", function()
      parser:parse("while true do\n x+1\ncatch\n x+2\nfinally\n x+3")
      assert(trans:next()):eq("while (true) try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)
  end):tags("while")

  -----------
  -- until --
  -----------
  suite("until", function()
    test("until Exp do Exp", function()
      parser:parse("until true do x+1")
      assert(trans:next()):eq("while (!(true)) {(x+1);}\n")
    end)

    test("until Exp do Stmt", function()
      parser:parse("until true do return x+1")
      assert(trans:next()):eq("while (!(true)) {return (x+1);}\n")
    end)

    test("until Exp do Block else Block", function()
      parser:parse("until true do x+1\nelse x+2")
      assert(trans:next()):eq("{let $__break__; while (!(true)) {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("until Decl; Exp do Sent", function()
      parser:parse("until x := 1; x == 10 do x + 1")
      assert(trans:next()):eq("for (let x=1; !((x==10)); ) {(x+1);}\n")
    end)

    test("until Decl; Exp do Block else Block", function()
      parser:parse("until x := 1; x == 10 do x + 2\nelse x + 3")
      assert(trans:next()):eq("{let $__break__; for (let x=1; !((x==10)); ) {(x+2);} if (!$__break__) {(x+3);}}\n")
    end)

    test("until with catch", function()
      parser:parse("until true do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("while (!(true)) try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("until Exp do Block else Block catch Block", function()
      parser:parse("until true do x+1\nelse x+2\ncatch\n x+3")
      assert(trans:next()):eq("try {let $__break__; while (!(true)) {(x+1);} if (!$__break__) {(x+2);}} catch(_) {(x+3);}\n")
    end)

    test("until with finally", function()
      parser:parse("until true do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("while (!(true)) try {(x+1);} finally {(x+2);}\n")
    end)

    test("until with catch and finally", function()
      parser:parse("until true do\n x+1\ncatch\n x+2\nfinally\n x+3")
      assert(trans:next()):eq("while (!(true)) try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)
  end):tags("until")

  --------
  -- do --
  --------
  suite("do", function()
    test("do with var", function()
      parser:parse("do var x, y\n x+=1\nwhile true")
      assert(trans:next()):eq("{let x, y;do {(x+=1);} while (true);}\n")
    end)

    test("do with while", function()
      parser:parse("do\n x+1\nwhile true")
      assert(trans:next()):eq("do {(x+1);} while (true);\n")
    end)

    test("do with catch", function()
      parser:parse("do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("do with finally", function()
      parser:parse("do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("try {(x+1);} finally {(x+2);}\n")
    end)

    test("do with catch and finally", function()
      parser:parse("do\n x+1\ncatch\n x+2\nfinally\n x+3")
        assert(trans:next()):eq("try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)

    test("do with while, catch and finally", function()
      parser:parse("do\n x+1\nwhile true\ncatch\n x+2\nfinally\n x+3")
        assert(trans:next()):eq("do try {(x+1);} catch(_) {(x+2);} finally {(x+3);} while (true);\n")
    end)
  end)

  ---------
  -- for --
  ---------
  suite("for", function()
    test("for Name ; Exp do Exp", function()
      parser:parse("for i ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i = 0; (i<10); ) {print(i);}\n")
    end)

    test("for Variable; Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do print(i)")
      assert(trans:next()):eq("for (let i = 0; (i<10); (i+=1)) {print(i);}\n")
    end)

    test("for Name, Name ; Exp do Exp", function()
      parser:parse("for i, j ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i, j; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0, j = 1 ; i < 10 do print(i)")
      assert(trans:next()):eq("for (let i = 0, j = 1; (i<10); ) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp; Exp do Exp", function()
      parser:parse("for i = 0, j = 1 ; i < 10; i += 1 do print(i)")
      assert(trans:next()):eq("for (let i = 0, j = 1; (i<10); (i+=1)) {print(i);}\n")
    end)

    test("for Name = Exp, Name = Exp ; Exp; Exp do Exp else Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do\n  print(i)\n  if i == 5 then break for\nelse\n  print(j)")
      assert(trans:next()):eq("{let $__break__; for (let i = 0; (i<10); (i+=1)) {print(i);if ((i==5)) {$__break__ = true; break;}} if (!$__break__) {print(j);}}\n")
    end)

    test("for Name; Exp; Exp do\\n Body\\ncatch\\n Body", function()
      parser:parse("for i; i < 10; i += 1 do\n print(i)\ncatch e\n print(e)")
      assert(trans:next()):eq("for (let i; (i<10); (i+=1)) try {print(i);} catch(e) {print(e);}\n")
    end)

    test("for Name; Exp; Exp do Block else Block catch Body", function()
      parser:parse("for i = 0; i < 10; i += 1 do\n print(i)\n  if i == 5 then break for\nelse\n  print(j)\ncatch e\n print(e)")
      assert(trans:next()):eq("try {let $__break__; for (let i = 0; (i<10); (i+=1)) {print(i);if ((i==5)) {$__break__ = true; break;}} if (!$__break__) {print(j);}} catch(e) {print(e);}\n")
    end)
  end):tags("for")

  --------------
  -- for each --
  --------------
  suite("for each", function()
    test("for each Name in Exp do\\n Sent", function(params)
      parser:parse("for each k " .. params[1] .. " arr do\n x+1")
      assert(trans:next()):eq("for (let k of arr) {(x+1);}\n")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # :=", params=":="}
    )

    test("for each Name ::= Exp do\\n Sent", function()
      parser:parse("for each k ::= arr do\n x+1")
      assert(trans:next()):eq("for (const k of arr) {(x+1);}\n")
    end)

    test("for each Name ::= Exp do Block else Block", function()
      parser:parse("for each k ::= arr do\n x+1\nelse\n x+2")
      assert(trans:next()):eq("{let $__break__; for (const k of arr) {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("for each Name, Name in Exp do\\n Sent", function()
      parser:parse("for each k, v in map do\n x+1")
      assert(trans:next()):eq("for (let [k, v] of Object.entries(map)) { {(x+1);} }\n")
    end)

    test("for each Name, Name in Exp do Block else Block", function()
      parser:parse("for each k, v in map do\n x+1\nelse\n x+2")
      assert(trans:next()):eq("{let $__break__; for (let [k, v] of Object.entries(map)) { {(x+1);} if (!$__break__) {(x+2);}}\n")
    end)

    test("for each with catch", function()
      parser:parse("for each k in arr do\n x+1\ncatch\n x+2")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} catch(_) {(x+2);}\n")
    end)

    test("for each with else and catch", function()
      parser:parse("for each k in arr do\n x+1\nelse\n x+2\ncatch\n x+3")
      assert(trans:next()):eq("try {let $__break__; for (let k of arr) {(x+1);} if (!$__break__) {(x+2);}} catch(_) {(x+3);}\n")
    end)

    test("for each with finally", function()
      parser:parse("for each k in arr do\n x+1\nfinally\n x+2")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} finally {(x+2);}\n")
    end)

    test("for each with catch and finally", function()
      parser:parse("for each k in arr do\n x+1\ncatch\n x+2\nfinally\n x+3")
      assert(trans:next()):eq("for (let k of arr) try {(x+1);} catch(_) {(x+2);} finally {(x+3);}\n")
    end)
  end):tags("foreach")

  ------------
  -- return --
  ------------
  suite("return", function()
    test("return", function()
      parser:parse("return")
      assert(trans:next()):eq("return;\n")
    end)

    test("return Exp", function()
      parser:parse("return x+1")
      assert(trans:next()):eq("return (x+1);\n")
    end)
  end)

  -----------
  -- yield --
  -----------
  suite("yield", function()
    test("yield Exp", function()
      parser:parse("yield x+1")
      assert(trans:next()):eq("yield (x+1);\n")
    end)
  end):tags("yield")

  ----------
  -- intf --
  ----------
  suite("intf", function()
    test("fields", function()
      parser:parse('intf Param\n  a\n  b?:text\n  c:num\n  d:(num,"text")\n  e?:(num,text)')
      assert(trans:next()):eq(
        "const Param = dogma.intf('Param', {a: {optional: false, type: null}, b: {optional: true, type: text}, c: {optional: false, type: num}, d: {optional: false, type: [num, \"text\"]}, e: {optional: true, type: [num, text]}});\n"
      )
    end)

    test("methods", function()
      parser:parse('intf Calcul\n  fn sum(x, y): num\n  fn sub(x, y): num')
      assert(trans:next()):eq(
        "const Calcul = dogma.intf('Calcul', {sum: {optional: false, type: func}, sub: {optional: false, type: func}});\n"
      )
    end)

    test("export intf", function()
      parser:parse("export intf Param\n  x")
      assert(trans:next()):eq(
        "const Param = dogma.intf('Param', {x: {optional: false, type: null}});module.exports = exports = Param;\n"
      )
    end)

    test("pub intf", function()
      parser:parse("pub intf Param\n  x")
      assert(trans:next()):eq(
        "const Param = dogma.intf('Param', {x: {optional: false, type: null}});module.exports['Param'] = exports['Param'] = Param;\n"
      )
    end)

    test("with base", function()
      parser:parse("intf Coord3D: Coord2D\n  z")
      assert(trans:next()):eq(
        "const Coord3D = dogma.intf('Coord3D', {z: {optional: false, type: null}}, Coord2D);\n"
      )
    end)

    test("with base but without members", function()
      parser:parse("intf Coord3D: Coord2D")
      assert(trans:next()):eq("const Coord3D = dogma.intf('Coord3D', {}, Coord2D);\n")
    end)
  end):tags("intf")

  ------------
  -- struct --
  ------------
  suite("struct", function()
    suite("template", function()
      test("newline expected", function()
        parser:parse("use Tmpl<T>\nexport struct MyStruct : Tmpl<MyT>var x = 123", "test/data/template/MyStruct.dog")
        assert(function() trans:next() end):raises("test/data/template/MyStruct.dog (2, 35): newline expected.")
      end)

      test("template file not found", function()
        parser:parse("use Tmpl<T>\nexport struct MyStruct : tmpl<MyT>", "test/data/template/MyStruct.dog")
        assert(function() trans:next() end):raises("test/data/template/MyStruct.dog (end of file): file not found: test/data/template/tmpl.dogt.")
      end)

      test("OK", function(params)
        parser:parse(
          string.format("use Tmpl<T>\n%s\nexport struct MyStruct : Tmpl<MyT>\nvar x = 123", params[1]),
          "test/data/template/MyStruct.dog"
        )

        assert(trans:next()):eq([[

const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('x', _['x'], null);
    Object.defineProperty(this, 'x', {value: coalesce(_['x'], null), writable: true, enumerable: true});
    expect('y', _['y'], null);
    Object.defineProperty(this, 'y', {value: coalesce(_['y'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_1858abefb5a2e787ec30ad1f05aa854c___init__ instanceof Function) this._pvt_1858abefb5a2e787ec30ad1f05aa854c___init__(_);
    /* istanbul ignore next */ if (this._pvt_1858abefb5a2e787ec30ad1f05aa854c___post__ instanceof Function) this._pvt_1858abefb5a2e787ec30ad1f05aa854c___post__();
    /* istanbul ignore next */ if (this._pvt_1858abefb5a2e787ec30ad1f05aa854c___validate__ instanceof Function) this._pvt_1858abefb5a2e787ec30ad1f05aa854c___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports = exports = MyStruct;
const Self = MyStruct;
MyStruct.prototype.inc = function(x) { /* istanbul ignore next */ dogma.expect("x", x, MyT);{} };
MyStruct.prototype.dec = function(x) { /* istanbul ignore next */ dogma.expect("x", x, MyT);{} };
let x = 123;
]])
      end):iter(
        {subtitle=" # with @Self", params="@Self"},
        {subtitle=" # with no @Self", params=""}
      )
    end)

    test("sealed", function()
      parser:parse("@sealed\nstruct Struct")
      assert(trans:next()):eq([[
const $Struct = class Struct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const Struct = new Proxy($Struct, { apply(receiver, self, args) { return Object.seal(new $Struct(...args)); } });
]])
    end)

    test("const", function()
      parser:parse("@const\nstruct Struct")
      assert(trans:next()):eq([[
const $Struct = class Struct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const Struct = new Proxy($Struct, { apply(receiver, self, args) { return dogma.immutable(new $Struct(...args)); } });
]])
    end)

    test("singleton", function()
      parser:parse("@singleton\nstruct Struct\n  @singleton\n  pub proc isCompatible(i, args)")
      assert(trans:next()):eq([[
const $Struct = class Struct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
    Object.defineProperty($Struct, '__SINGLETON__', {value: this, writable: false, enumerable: false});
  }
};
const Struct = new Proxy($Struct, { apply(receiver, self, args) { const i = $Struct.__SINGLETON__; if (!i) return new $Struct(...args); if (Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___singleton__ instanceof Function) Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___singleton__(i, args); return i; } });
Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___singleton__ = Struct.isCompatible = function(i, args) { /* istanbul ignore next */ dogma.expect("i", i);/* istanbul ignore next */ dogma.expect("args", args);{} };
]])
    end)

    test("factory", function()
      parser:parse("@factory\nstruct Struct\n  @factory\n  pvt fn create(args) = new(args)")
      assert(trans:next()):eq([[
const $Struct = class Struct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const Struct = new Proxy($Struct, { apply(receiver, self, args) { return Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___factory__(...args); } });
{
const $$Class$$ = Struct;
Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___factory__ = Struct._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_create = function(args) { /* istanbul ignore next */ dogma.expect("args", args);{return new $$Class$$(args);} };
}
]])
    end)

    test("struct Name", function()
      parser:parse("struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@Self struct Name", function()
      parser:parse("@Self struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
const Self = MyStruct;
]])
    end)

    test("@emitter struct Name", function()
      parser:parse("@emitter struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct extends dogma.EventEmitter {
  constructor(_) {
    super();
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@abstract struct Name", function()
      parser:parse("@abstract struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'MyStruct' is abstract."); } });
]])
    end)

    test("export struct Name", function()
      parser:parse("export struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports = exports = MyStruct;
]])
    end)

    test("export struct Name :: Name", function()
      parser:parse("export struct A :: B")
      assert(trans:next()):eq([[
const $A = class A {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
module.exports = exports = A;
]])
    end)

    test("export struct Name :: Name, Name", function()
      parser:parse("export struct A :: B, C")
      assert(trans:next()):eq([[
const $A = class A {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(_);/* istanbul ignore next */ if (C.prototype._constructor instanceof Function) C.prototype._constructor.bind(this)(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};dogma.mixin($A, B);dogma.mixin($A, C);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
module.exports = exports = A;
]])
    end)

    test("pub struct Name", function()
      parser:parse("pub struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
module.exports['MyStruct'] = exports['MyStruct'] = MyStruct;
]])
    end)

    test("struct Name : Name", function()
      parser:parse("struct MyStruct : MyBase")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct extends MyBase {
  constructor(_) {
    super(_);
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@mixable struct Name", function()
      parser:parse("@mixable struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor() {
    this._constructor(...arguments);
  }

  _constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("@mixable @abstract struct Name", function()
      parser:parse("@mixable @abstract struct MyStruct")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  /*istanbul ignore next*/
  constructor() {
    this._constructor(...arguments);
  }

  /*istanbul ignore next*/
  _constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'MyStruct' is abstract."); } });
]])
    end)

    test("@defer struct Name", function()
      parser:parse("@defer\nstruct MyStruct\n  pub proc method1()\n    1+2\n    self.defer(3+4)\n  pub proc method2()\n    1+2\n    self.deferred()")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    Object.defineProperty(this, '$deferred', {value: [], writable: false, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
MyStruct.prototype.method1 = function() { {(1+2);this.$deferred.push(() => (3+4));} };
MyStruct.prototype.method2 = function() { {(1+2);{const errors = []; while (this.$deferred.length > 0) {try {const r = this.$deferred.pop()(); /*istanbul ignore next*/if (r instanceof Promise) r.then(() => null, () => null);} catch(e) {errors.push(e);}} /*istanbul ignore next*/if (errors.length > 0) throw new Error(errors); };} };
]])
    end)

    suite("list fields", function()
      test("type[...]", function()
        parser:parse("struct MyStruct\n  pub var a: num[]\n  pub var b: true[2]\n  pub var c: text[2..*]\n  pub var d: text[2..4]")
        assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], dogma.TypeDef({name: 'inline', types: [num], min: 0, max: null}));
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: true, enumerable: true});
    expect('b', _['b'], dogma.TypeDef({name: 'inline', types: [true], min: 2, max: 2}));
    Object.defineProperty(this, 'b', {value: coalesce(_['b'], null), writable: true, enumerable: true});
    expect('c', _['c'], dogma.TypeDef({name: 'inline', types: [text], min: 2, max: null}));
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: true, enumerable: true});
    expect('d', _['d'], dogma.TypeDef({name: 'inline', types: [text], min: 2, max: 4}));
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
      end)

      test("(types)[...]", function()
        parser:parse("struct MyStruct\n  pub var a: (m,n)[]\n  pub var b: (m,n)[2]\n  pub var c: (m,n)[2..*]\n  pub var d: (m,n)[2..4]")
        assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], dogma.TypeDef({name: 'inline', types: [m, n], min: 0, max: null}));
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: true, enumerable: true});
    expect('b', _['b'], dogma.TypeDef({name: 'inline', types: [m, n], min: 2, max: 2}));
    Object.defineProperty(this, 'b', {value: coalesce(_['b'], null), writable: true, enumerable: true});
    expect('c', _['c'], dogma.TypeDef({name: 'inline', types: [m, n], min: 2, max: null}));
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: true, enumerable: true});
    expect('d', _['d'], dogma.TypeDef({name: 'inline', types: [m, n], min: 2, max: 4}));
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
      end)
    end)

    test("field: name: type(type)", function()
      parser:parse("struct MyStruct\n  pub var a: num(text)")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], [num, text]);
    Object.defineProperty(this, 'a', {value: (dogma.is(_['a'], num) ? _['a'] : num(_['a'])), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("field: name?: type(type)", function()
      parser:parse("struct MyStruct\n  pub var a?: num(text)")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (_['a'] != null) expect('a', _['a'], [num, text]);
    Object.defineProperty(this, 'a', {value: (_['a'] != null ? (dogma.is(_['a'], num) ? _['a'] : num(_['a'])) : null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("var fields", function()
      parser:parse("struct MyStruct\n  pub var a\n  intl var b?\n  pub var c:num\n  pub var d?:num\n  pub var e:=123\n  pub var f:text=123\n  pub var g=123\n  pub var h=_.h")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: true, enumerable: true});
    Object.defineProperty(this, '_b', {value: coalesce(_['b'], null), writable: true, enumerable: false});
    expect('c', _['c'], num);
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (_['d'] != null) expect('d', _['d'], num);
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], null), writable: true, enumerable: true});
    /* istanbul ignore next */ if (_['e'] != null) expect('e', _['e'], num);
    Object.defineProperty(this, 'e', {value: coalesce(_['e'], 123), writable: true, enumerable: true});
    /* istanbul ignore next */ if (_['f'] != null) expect('f', _['f'], text);
    Object.defineProperty(this, 'f', {value: coalesce(_['f'], 123), writable: true, enumerable: true});
    Object.defineProperty(this, 'g', {value: coalesce(_['g'], 123), writable: true, enumerable: true});
    Object.defineProperty(this, 'h', {value: coalesce(_['h'], _.h), writable: true, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("fields with @strict", function()
      parser:parse("struct MyStruct\n  @strict pub var a=123\n  @strict intl const b=321")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    Object.defineProperty(this, 'a', {value: 123, writable: true, enumerable: true});
    Object.defineProperty(this, '_b', {value: 321, writable: false, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("fields with @prop", function()
      parser:parse("struct MyStruct\n  @prop @strict pub var a=123\n  @prop @strict intl const b=321")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    Object.defineProperty(this, 'a', {value: 123, writable: true, enumerable: true});
    Object.defineProperty(this, '_b', {value: 321, writable: false, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
Object.defineProperty(MyStruct.prototype, 'b', {get() { return this._b; }});
]])
    end)

    test("fields with @static", function()
      parser:parse("@Self\nstruct MyStruct\n  @static pub const DEFAULT = Self()")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
const Self = MyStruct;
Object.defineProperty(MyStruct, 'DEFAULT', {value: Self(), writable: false, enumerable: true});
]])
    end)

    test("const fields", function()
      parser:parse("struct MyStruct\n  pub const a\n  intl const b=text\n  pub const c:text\n  pub const d:text=num\n  pub const e:=123\n  pub const f?:text")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: false, enumerable: true});
    Object.defineProperty(this, '_b', {value: coalesce(_['b'], text), writable: false, enumerable: false});
    expect('c', _['c'], text);
    Object.defineProperty(this, 'c', {value: coalesce(_['c'], null), writable: false, enumerable: true});
    /* istanbul ignore next */ if (_['d'] != null) expect('d', _['d'], text);
    Object.defineProperty(this, 'd', {value: coalesce(_['d'], num), writable: false, enumerable: true});
    /* istanbul ignore next */ if (_['e'] != null) expect('e', _['e'], num);
    Object.defineProperty(this, 'e', {value: coalesce(_['e'], 123), writable: false, enumerable: true});
    /* istanbul ignore next */ if (_['f'] != null) expect('f', _['f'], text);
    Object.defineProperty(this, 'f', {value: coalesce(_['f'], null), writable: false, enumerable: true});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("public fields", function()
      parser:parse("struct MyStruct\n  pub var a\n  @hidden pub var b")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, 'a', {value: coalesce(_['a'], null), writable: true, enumerable: true});
    expect('b', _['b'], null);
    Object.defineProperty(this, 'b', {value: coalesce(_['b'], null), writable: true, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("private fields", function()
      parser:parse("struct MyStruct\n  pvt var a\n  pvt const b")
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    expect('a', _['a'], null);
    Object.defineProperty(this, '_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_a', {value: coalesce(_['a'], null), writable: true, enumerable: false});
    expect('b', _['b'], null);
    Object.defineProperty(this, '_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_b', {value: coalesce(_['b'], null), writable: false, enumerable: false});
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
]])
    end)

    test("events", function()
      parser:parse([[
        @emitter
        struct MyStruct
          pub event start
          pub event stop: {task: Task, ts: timestamp}
      ]])
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct extends dogma.EventEmitter {
  constructor(_) {
    super();
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
MyStruct.__events__ = {};
MyStruct.__events__["start"] = {type: null};
MyStruct.__events__["stop"] = {type: dogma.intf("inline", {task: {optional: false, type: Task}, ts: {optional: false, type: timestamp}})};
]])
    end)

    test("struct methods", function()
      parser:parse([[
        struct MyStruct
          @init
          pvt fn method0(props)

          @validate
          pub fn method1()

          @post
          pvt proc method2()

          @post
          intl proc method3()

          @abstract
          pub proc method4()

          @catch
          pub proc method5()

          @catch
          pub fn method6()

          @static @catch
          pub proc method7()

          @static @catch
          pub fn method8()

          @overload
          pub fn method9()

          @overload
          pub proc method10()
      ]])
      assert(trans:next()):eq([[
const $MyStruct = class MyStruct {
  constructor(_) {
    /* istanbul ignore next */ if (_ == null) _ = {};
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__(_);
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__();
    /* istanbul ignore next */ if (this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function) this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__();
  }
};
const MyStruct = new Proxy($MyStruct, { apply(receiver, self, args) { return new $MyStruct(...args); } });
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method0 = function(props) { /* istanbul ignore next */ dogma.expect("props", props);{} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ = MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method0;
MyStruct.prototype.method1 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ = MyStruct.prototype.method1;
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method2 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ = MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_method2;
MyStruct.prototype._method3 = function() { {} };
MyStruct.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ = MyStruct.prototype._method3;
/* istanbul ignore next */ MyStruct.prototype.method4 = function() { abstract(); };
MyStruct.prototype.method5 = function() { try { {} } catch { } };
MyStruct.prototype.method6 = function() { try { {} } catch { return null; } };
MyStruct.method7 = function() { try { {} } catch { } };
MyStruct.method8 = function() { try { {} } catch { return null; } };


]])
    end)

    test("??", function()
      parser:parse('struct MyStruct\n  @strict pub var a = _.a ?? "hi"\n  pub var b = _.b ?? "bye"')
      assert(trans:next()):like([[
const $MyStruct = class MyStruct {
  constructor%(_%) {
    /%* istanbul ignore next %*/ if %(_ == null%) _ = {};
    let $$aux%d+;Object%.defineProperty%(this, 'a', {value: %($$aux%d+ = _%.a, %($$aux%d+ == null %? "hi" : $$aux%d+%)%), writable: true, enumerable: true}%);
    let $$aux%d+;Object%.defineProperty%(this, 'b', {value: coalesce%(_%['b'%], %($$aux%d+ = _%.b, %($$aux%d+ == null %? "bye" : $$aux%d+%)%)%), writable: true, enumerable: true}%);
    /%* istanbul ignore next %*/ if %(this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__ instanceof Function%) this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___init__%(_%);
    /%* istanbul ignore next %*/ if %(this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__ instanceof Function%) this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___post__%(%);
    /%* istanbul ignore next %*/ if %(this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__ instanceof Function%) this%._pvt_e6eace38db3ac6dfd7468b60ae7a70ae___validate__%(%);
  }
};
const MyStruct = new Proxy%($MyStruct, { apply%(receiver, self, args%) { return new $MyStruct%(%.%.%.args%); } }%);
]])
    end)
  end):tags("struct")

  -------------
  -- typedef --
  -------------
  suite("typedef", function()
    test("type A = B", function()
      parser:parse("type A = B")
      assert(trans:next()):eq(
        "const A = B;\n"
      )
    end)

    test("type A = (B,C)", function()
      parser:parse("type A = (B,C)")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B,C], min: null, max: null});\n"
      )
    end)

    test("type A = B[]", function()
      parser:parse("type A = B[]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 0, max: null});\n"
      )
    end)

    test("type A = B[2]", function()
      parser:parse("type A = B[2]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 2, max: 2});\n"
      )
    end)

    test("type A = B[2..*]", function()
      parser:parse("type A = B[2..*]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 2, max: null});\n"
      )
    end)

    test("type A = B[2..4]", function()
      parser:parse("type A = B[2..4]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 2, max: 4});\n"
      )
    end)

    test("export type A = B", function()
      parser:parse("export type A = B[2..4]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 2, max: 4});module.exports = exports = A;\n"
      )
    end)

    test("pub type A = B", function()
      parser:parse("pub type A = B[2..4]")
      assert(trans:next()):eq(
        "const A = dogma.TypeDef({name: 'A', types: [B], min: 2, max: 4});module.exports['A'] = exports['A'] = A;\n"
      )
    end)
  end):tags("typedef")

  ----------
  -- type --
  ----------
  suite("type", function()
    test("@emitter type Name()", function()
      parser:parse("@emitter type Name()")
      assert(trans:next()):eq([[
const $Name = class Name extends dogma.EventEmitter {
  constructor() { super();{}  }
};
const Name = new Proxy($Name, { apply(receiver, self, args) { return new $Name(...args); } });
]])
    end)

    test("@todo type Name()", function()
      parser:parse("@todo type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { todo(); }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@Self type Name()", function()
      parser:parse("@Self type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
const Self = Coord2D;
]])
    end)

    test("@component type Name", function()
      parser:parse("@component type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { dogma.autobind(this, $Coord2D);/* istanbul ignore next */if ($obj == null) $obj = {};  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@Self type Name", function()
      parser:parse("@Self type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
const Self = Coord2D;
]])
    end)

    test("type Name", function()
      parser:parse("type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@abstract export type Name", function()
      parser:parse("@abstract\nexport type Coord2D")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'Coord2D' is abstract."); } });module.exports = exports = Coord2D;
]])
    end)

    test("type Name with fields", function()
      parser:parse('type Coord3D\n  const x\n  var y?\n  var z?:num\n  pvt const kind="3D"')
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};/* istanbul ignore next */ dogma.expect("x", $obj["x"], null);Object.defineProperty(this, "x", {value: $obj["x"], enum: true, writable: false});Object.defineProperty(this, "y", {value: $obj["y"], enum: true, writable: true});/* istanbul ignore next */if ($obj["z"] != null) dogma.expect("z", $obj["z"], num);Object.defineProperty(this, "z", {value: $obj["z"], enum: true, writable: true});/* istanbul ignore next */if ($obj["kind"] != null) dogma.expect("kind", $obj["kind"], null);Object.defineProperty(this, "_kind", {value: $obj["kind"]||"3D", writable: false});  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Name()\\n Body", function()
      parser:parse("type Coord2D()\n .x = 0\n .y = 0")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {(this.x=0);(this.y=0);}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@component type Name()\\n Body", function()
      parser:parse("@component\ntype Coord2D()\n .x = 0\n .y = 0")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { dogma.autobind(this, $Coord2D);{(this.x=0);(this.y=0);}  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("type Coord3D(.x, :y, !z)", function()
      parser:parse("type Coord2D(.x, :y, !z)")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);/* istanbul ignore next */ dogma.expect("z", z);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "_y", {value: y, writable: true});Object.defineProperty(this, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_z", {value: z, writable: true});{}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("@component type Coord2D(.x, .y)", function()
      parser:parse("@component type Coord2D(.x, .y)")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);dogma.autobind(this, $Coord2D);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{}  }
};Object.defineProperty($Coord2D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
    end)

    test("type Name(:Name)", function()
      parser:parse("type Coord1D(:x)")
      assert(trans:next()):eq([[
const $Coord1D = class Coord1D {
  constructor(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "_x", {value: x, writable: true});{}  }
};
const Coord1D = new Proxy($Coord1D, { apply(receiver, self, args) { return new $Coord1D(...args); } });
]])
    end)

    test("type Name() : Name", function()
      parser:parse("type Coord3D() : Coord2D")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { {}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Name() : Name.Name()", function()
      parser:parse("type Coord3D() : Coord.D2()")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord.D2 {
  constructor() { super();{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(x, y, .z) : Coord2D(x, y)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(x, y)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);/* istanbul ignore next */ dogma.expect("z", z);super(x, y);Object.defineProperty(this, "z", {value: z, enum: true, writable: true});{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(x, y, .z) : Coord2D(...)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor(x, y, z) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);/* istanbul ignore next */ dogma.expect("z", z);super(...arguments);Object.defineProperty(this, "z", {value: z, enum: true, writable: true});{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("@component type Coord3D(...) : Coord2D(...)", function()
      parser:parse("@component type Coord3D(...) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { super(...arguments);dogma.autobind(this, $Coord3D);{}  }
};Object.defineProperty($Coord3D, '__bind__', {value: [], enumerable: false, writable: false});
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("type Coord3D(...) : Coord2D(...)", function()
      parser:parse("type Coord3D(...) : Coord2D(...)")
      assert(trans:next()):eq([[
const $Coord3D = class Coord3D extends Coord2D {
  constructor() { super(...arguments);{}  }
};
const Coord3D = new Proxy($Coord3D, { apply(receiver, self, args) { return new $Coord3D(...args); } });
]])
    end)

    test("export type Coord2D()", function()
      parser:parse("export type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });module.exports = exports = Coord2D;
]])
    end)

    test("@abstract export type Coord2D()", function()
      parser:parse("@abstract\nexport type Coord2D()")
      assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { {}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { /* istanbul ignore next */ throw("'Coord2D' is abstract."); } });module.exports = exports = Coord2D;
]])
    end)

    suite("mixin", function()
      test("@mixable type Name()", function()
        parser:parse("@mixable type Coord2D(x, y)")
        assert(trans:next()):eq([[
const $Coord2D = class Coord2D {
  constructor() { this._constructor(...arguments); } _constructor(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{}  }
};
const Coord2D = new Proxy($Coord2D, { apply(receiver, self, args) { return new $Coord2D(...args); } });
]])
      end)

      test("type Name() :: Mixin()\\n Body", function()
        parser:parse("type A() :: B()\n .x = 0\n .y = 0")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();{(this.x=0);(this.y=0);}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name :: Mixin()", function()
        parser:parse("type A :: B()")
        assert(trans:next()):eq([[
const $A = class A {
  constructor($obj) { /* istanbul ignore next */if ($obj == null) $obj = {};/* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin()", function()
        parser:parse("type A() :: B()")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)();{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin(...)", function()
        parser:parse("type A() :: B(...)")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(...arguments);{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)

      test("type Name() :: Mixin(args)", function()
        parser:parse("type A() :: B(one, two)")
        assert(trans:next()):eq([[
const $A = class A {
  constructor() { /* istanbul ignore next */ if (B.prototype._constructor instanceof Function) B.prototype._constructor.bind(this)(one, two);{}  }
};dogma.mixin($A, B);
const A = new Proxy($A, { apply(receiver, self, args) { return new $A(...args); } });
]])
      end)
    end):tags("mixin")
  end):tags("type")

  ----------
  -- proc --
  ----------
  suite("proc", function()
    test("@overload", function()
      parser:parse("@overload\nproc p(x: num)")
      assert(trans:next()):eq('\n')
    end)

    suite("@input", function()
      test("procedure", function()
        parser:parse("@input\nproc f(x) x")
        assert(trans:next()):eq('function f(x) { print(...arguments);/* istanbul ignore next */ dogma.expect("x", x);{x;} }\n')
      end)

      test("method", function()
        parser:parse("@input\npr Type.f(x) x")
        assert(trans:next()):eq('Type.prototype.f = function(x) { print(...arguments);/* istanbul ignore next */ dogma.expect("x", x);{x;} };\n')
      end)
    end)

    suite("@bind", function()
      test("procedure", function()
        parser:parse("@bind\npr sum(x?, y?) x + y")
        assert(trans:next()):eq("function sum(x, y) { {(x+y);} }\n")
      end)

      test("method", function()
        parser:parse("@bind\nproc Calcul.sum(x?, y?) x + y")
        assert(trans:next()):eq("Calcul.__bind__.push(['sum', function(x, y) { {(x+y);} }]);\n")
      end)
    end)

    suite("standalone", function()
      test("@todo proc Name()", function()
        parser:parse("@todo\nproc myfn()")
        assert(trans:next()):eq("function myfn() { todo(); }\n")
      end)

      test("proc Name()", function()
        parser:parse("proc myfn()")
        assert(trans:next()):eq("function myfn() { {} }\n")
      end)

      test("@catch", function()
        parser:parse('@catch\nproc myproc()\n  x + y')
        assert(trans:next()):eq("function myproc() { try { {(x+y);} } catch { } }\n")
      end)

      suite("@alias", function()
        test("@alias proc Name()", function()
          parser:parse('@alias("myfn2") proc myfn()')
          assert(trans:next()):eq("function myfn() { {} } const myfn2 = myfn;\n")
        end)

        test("@alias pub proc Name()", function()
          parser:parse('@alias("myfn2") pub proc myfn()')
          assert(trans:next()):eq("function myfn() { {} }export {myfn}; const myfn2 = myfn;export {myfn2};\n")
        end)
      end)

      test("proc Name() Exp", function()
        parser:parse("proc myfn() 123")
        assert(trans:next()):eq("function myfn() { {123;} }\n")
      end)

      test("export proc Name()", function()
        parser:parse("export proc myfn()")
        assert(trans:next()):eq('function myfn() { {} }module.exports = exports = myfn;\n')
      end)

      test("pub proc Name()", function()
        parser:parse("pub proc myfn()")
        assert(trans:next()):eq('function myfn() { {} }export {myfn};\n')
      end)

      test("proc Name(params) Body", function()
        parser:parse("proc sum(x, y) x+y")
        assert(trans:next()):eq('function sum(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{(x+y);} }\n')
      end)

      test("async proc Name(params) Body", function()
        parser:parse("async proc sum(x, y) x+y")
        assert(trans:next()):eq('async function sum(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{(x+y);} }\n')
      end)
    end)

    suite("type", function()
      suite("static method", function()
        test("@static proc Name . Name() Body", function()
          parser:parse("@static\nproc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = function() { {(1+2);} };\n")
        end)

        test("@static async proc Name . Name() Body", function()
          parser:parse("@static\nasync proc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = async function() { {(1+2);} };\n")
        end)

        test("@static @todo proc Name . Name() Body", function()
          parser:parse("@static @todo\nproc MyType.method() 1+2")
          assert(trans:next()):eq("MyType.method = function() { todo(); };\n")
        end)
      end):tags("static")

      suite("@alias", function()
        test("@alias proc Name . Name ()", function()
          parser:parse('@alias("myfn2") proc MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.myfn2 = MyType.prototype.myfn = function() { {} };\n')
        end)
      end)

      test("proc Name . Name ()", function()
        parser:parse("proc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = function() { {} };\n')
      end)

      test("async proc Name . Name ()", function()
        parser:parse("async proc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = async function() { {} };\n')
      end)

      suite("abstract", function()
        test("pub", function()
          parser:parse("@abstract\nproc MyType.myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype.myfn = function() { abstract(); };\n')
        end)

        test("intl", function()
          parser:parse("@abstract\nproc MyType:myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype._myfn = function() { abstract(); };\n')
        end)

        test("pvt", function()
          parser:parse("@abstract\nproc MyType!myfn()")
          assert(function() trans:next() end):raises("anonymous code (2, 1): abstract method can't be private (!).")
        end)
      end):tags("abstract")

      test("@todo\\nproc Name . Name ()", function()
        parser:parse("@todo\nproc MyType.myfn()")
        assert(trans:next()):eq('MyType.prototype.myfn = function() { todo(); };\n')
      end)

      test("proc Name : Name ()", function()
        parser:parse("proc MyType:myfn()")
        assert(trans:next()):eq('MyType.prototype._myfn = function() { {} };\n')
      end)

      test("proc Name ! Name ()", function()
        parser:parse("proc MyType!myfn()")
        assert(trans:next()):eq('MyType.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_myfn = function() { {} };\n')
      end)

      test("proc Name . Name (. Name)", function()
        parser:parse("proc MyType.myfn(.x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
      end)

      test("proc Name . Name (. Name , . Name)", function()
        parser:parse("proc MyType.myfn(.x, .y)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{} };\n')
      end)

      test("proc Name . Name (const . Name)", function()
        parser:parse("proc MyType.myfn(const .x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
      end)

      test("proc Name . Name (const . Name)", function()
        parser:parse("proc MyType.myfn(const .x)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
      end)

      test("proc Name . Name (.x Name ?)", function()
        parser:parse("proc MyType.myfn(.x?)")
        assert(trans:next()):eq('MyType.prototype.myfn = function(x) { Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
      end)
    end)
  end):tags("proc")

  --------
  -- fn --
  --------
  suite("fn", function()
    test("@overload", function()
      parser:parse("@overload\nfn sum(x: num, y: num)")
      assert(trans:next()):eq('\n')
    end)

    suite("@input", function()
      test("function", function()
        parser:parse("@input\nfn f(x) = x")
        assert(trans:next()):eq('function f(x) { print(...arguments);/* istanbul ignore next */ dogma.expect("x", x);{return x;} }\n')
      end)

      test("method", function()
        parser:parse("@input\nfn Type.f(x) = x")
        assert(trans:next()):eq('Type.prototype.f = function(x) { print(...arguments);/* istanbul ignore next */ dogma.expect("x", x);{return x;} };\n')
      end)
    end)

    suite("@bind", function()
      test("function", function()
        parser:parse("@bind\nfn sum(x?, y?) = x + y")
        assert(trans:next()):eq("function sum(x, y) { {return (x+y);} }\n")
      end)

      test("method", function()
        parser:parse("@bind\nfn Calcul.sum(x?, y?) = x + y")
        assert(trans:next()):eq("Calcul.__bind__.push(['sum', function(x, y) { {return (x+y);} }]);\n")
      end)
    end)

    suite("params", function()
      suite("unpack", function()
        suite("=>", function()
          test("fn(params=>[...])", function()
            parser:parse("fn sum(params=>[x,y]) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.expect("params", params, list);let [x,y] = params;{return (x+y);} }\n')
          end)

          test("fn(params=>{...})", function()
            parser:parse("fn sum(params=>{x,y}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.expect("params", params, map);let {x,y} = params;{return (x+y);} }\n')
          end)

          test("fn(params:{...}=>{...})", function()
            parser:parse("fn sum(params:{x:num,y:num,z:num}=>{x,y}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.expect(\"params\", params, dogma.intf("inline", {x: {optional: false, type: num}, y: {optional: false, type: num}, z: {optional: false, type: num}}));let {x,y} = params;{return (x+y);} }\n')
          end)
        end)

        suite(":>", function()
          test("fn(params:> {x: num, y: num})", function()
            parser:parse("fn sum(params:> {x: num, y: num}) = x+y")
            assert(trans:next()):eq('function sum(params) { /* istanbul ignore next */ dogma.expect(\"params\", params, dogma.intf(\"inline\", {x: {optional: false, type: num}, y: {optional: false, type: num}}));let {x,y} = params;{return (x+y);} }\n')
          end)

          test("fn(p1:> {x: num(bool)}, p2:> {y?: num(text)})", function()
            parser:parse("fn sum(p1:> {x: num(bool)}, p2:> {y?: num(text)}) = x+y")
            assert(trans:next()):eq('function sum(p1, p2) { /* istanbul ignore next */ dogma.expect("p1", p1, dogma.intf("inline", {x: {optional: false, type: [num, bool]}}));/* istanbul ignore next */ dogma.expect("p2", p2, dogma.intf("inline", {y: {optional: true, type: [num, text]}}));let {x} = p1;if (!dogma.is(x, num)) x = num(x);let {y} = p2;if (y != null && !dogma.is(y, num)) y = num(y);{return (x+y);} }\n')
          end)

          test("fn(params?:> {x: num(bool)})", function()
            parser:parse("fn number(params?:> {x: num(bool)}) = x")
            assert(trans:next()):eq('function number(params) { /* istanbul ignore next */ if (params != null) dogma.expect("params", params, dogma.intf("inline", {x: {optional: false, type: [num, bool]}}));let {x} = params || {};if (!dogma.is(x, num)) x = num(x);{return x;} }\n')
          end)

          test("fn(params?:> {x?: num(bool)})", function()
            parser:parse("fn number(params?:> {x?: num(bool)}) = x")
            assert(trans:next()):eq('function number(params) { /* istanbul ignore next */ if (params != null) dogma.expect("params", params, dogma.intf("inline", {x: {optional: true, type: [num, bool]}}));let {x} = params || {};if (x != null && !dogma.is(x, num)) x = num(x);{return x;} }\n')
          end)
        end)
      end)

      test("@noParamCheck", function()
        parser:parse("@noParamCheck fn f(param: num)")
        assert(trans:next()):eq('function f(param) { {} }\n')
      end)

      test("fn Name(keyword)", function()
        parser:parse("fn f(class)")
        assert(trans:next()):eq('function f(class_) { /* istanbul ignore next */ dogma.expect("class_", class_);{} }\n')
      end)

      test("fn Name(Name)", function()
        parser:parse("fn f(x)")
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x);{} }\n')
      end)

      test("fn Name(Name:Name)", function()
        parser:parse("fn f(x:num)")
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, num);{} }\n')
      end)

      test('fn Name(Name:Text)', function()
        parser:parse('fn f(x:"num")')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, "num");{} }\n')
      end)

      test('fn Name(Name:Number)', function()
        parser:parse('fn f(x:123)')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, 123);{} }\n')
      end)

      test('fn Name(Name:true)', function()
        parser:parse('fn f(x:true)')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, true);{} }\n')
      end)

      test('fn Name(Name:false)', function()
        parser:parse('fn f(x:false)')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, false);{} }\n')
      end)

      test("fn Name(Name:(One))", function()
        parser:parse('fn f(x:(text))')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, [text]);{} }\n')
      end)

      test("fn Name(Name:(Name, Text))", function()
        parser:parse('fn f(x:(text,"list"))')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, [text, "list"]);{} }\n')
      end)

      test("fn Name(Name: list(Name))", function()
        parser:parse("fn f(x: list(text))")
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, [list, text]);if (!dogma.is(x, list)) x = list(x);{} }\n')
      end)

      test("fn Name(Name?: list(Name))", function()
        parser:parse("fn f(x?: list(text))")
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, [list, text]);if (x != null && !dogma.is(x, list)) x = list(x);{} }\n')
      end)

      test("fn Name(Name: list(Name, Text))", function()
        parser:parse('fn f(x: list(text, "hello"))')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ dogma.expect("x", x, [list, text, "hello"]);if (!dogma.is(x, list)) x = list(x);{} }\n')
      end)

      test("fn Name(Name?: list(Name, Text))", function()
        parser:parse('fn f(x?: list(text, "hello"))')
        assert(trans:next()):eq('function f(x) { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, [list, text, "hello"]);if (x != null && !dogma.is(x, list)) x = list(x);{} }\n')
      end)

      test("fn Name(const Name)", function()
        parser:parse("fn myfn(x)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ dogma.expect("x", x);{} }\n')
      end)

      test("fn Name(Name?)", function()
        parser:parse("fn myfn(x?)")
        assert(trans:next()):eq('function myfn(x) { {} }\n')
      end)

      test("fn Name(Name?:Name)", function()
        parser:parse("fn myfn(x?:num)")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, num);{} }\n')
      end)

      test("fn Name(Name?:(Name))", function()
        parser:parse("fn myfn(x?:(num))")
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, [num]);{} }\n')
      end)

      test("fn Name(Name?:(One, Two))", function()
        parser:parse('fn myfn(x?:(num, "num"))')
        assert(trans:next()):eq('function myfn(x) { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, [num, "num"]);{} }\n')
      end)

      test("fn Name(...Name)", function()
        parser:parse("fn myfn(...x)")
        assert(trans:next()):eq('function myfn(...x) { {} }\n')
      end)

      test("fn Name(Name=Name)", function()
        parser:parse("fn myfn(x=123)")
        assert(trans:next()):eq('function myfn(x = 123) { {} }\n')
      end)

      test("fn Name(Name:{})", function()
        parser:parse("fn sum(vals:{}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ dogma.expect("vals", vals, dogma.intf("inline", {}));{return (vals.x+vals.y);} }\n')
      end)

      test("fn Name(Name:{p,q:type,r?:type,s:(type1,type2))", function()
        parser:parse("fn sum(vals:{p, q:num, r?:num, s:(text,num)}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ dogma.expect("vals", vals, dogma.intf("inline", {p: {optional: false, type: null}, q: {optional: false, type: num}, r: {optional: true, type: num}, s: {optional: false, type: [text, num]}}));{return (vals.x+vals.y);} }\n')
      end)

      test("fn Name(Name?:{p,q:type,r?:type,s:(type1,type2))", function()
        parser:parse("fn sum(vals?:{p, q:num, r?:num, s:(text,num)}) = vals.x + vals.y")
        assert(trans:next()):eq('function sum(vals) { /* istanbul ignore next */ if (vals != null) dogma.expect("vals", vals, dogma.intf("inline", {p: {optional: false, type: null}, q: {optional: false, type: num}, r: {optional: true, type: num}, s: {optional: false, type: [text, num]}}));{return (vals.x+vals.y);} }\n')
      end)
    end)

    suite("return var", function()
      test("fn Name() -> self Body", function()
        parser:parse("fn myfn() -> self\n 1+2")
        assert(trans:next()):eq("function myfn() { {(1+2);} return this; }\n")
      end)

      test("fn Name() -> Name Body", function()
        parser:parse("fn myfn() -> x\n 1+2")
        assert(trans:next()):eq("function myfn() { let x;{(1+2);} return x; }\n")
      end)

      test("fn Name(param) -> Name Body - return variable not existing", function()
        parser:parse("fn myfn(a?, b?) -> x\n x = a + b")
        assert(trans:next()):eq("function myfn(a, b) { let x;{(x=(a+b));} return x; }\n")
      end)

      test("fn Name(param) -> Name Body - return variable existing", function()
        parser:parse("fn myfn(a?, b?, x?) -> x\n x = a + b")
        assert(trans:next()):eq("function myfn(a, b, x) { {(x=(a+b));} return x; }\n")
      end)

      test("fn Name() -> res:map", function()
        parser:parse("fn sum(x?, y?) -> res:map\n res.result = x + y")
        assert(trans:next()):eq("function sum(x, y) { let res = {};{(res.result=(x+y));} return res; }\n")
      end)

      test("fn Name() -> res:list", function()
        parser:parse("fn sum(x?, y?) -> res:list\n res.push(x + y)")
        assert(trans:next()):eq("function sum(x, y) { let res = [];{res.push((x+y));} return res; }\n")
      end)

      test("fn Name() -> res:set", function()
        parser:parse("fn sum(x?, y?) -> res:set\n res.add(x + y)")
        assert(trans:next()):eq("function sum(x, y) { let res = set();{res.add((x+y));} return res; }\n")
      end)

      test("fn Name() -> res:bool", function()
        parser:parse("fn sum(x?, y?) -> res:bool")
        assert(trans:next()):eq("function sum(x, y) { let res = false;{} return res; }\n")
      end)

      test("fn Name() -> res:text", function()
        parser:parse("fn sum(x?, y?) -> res:text")
        assert(trans:next()):eq('function sum(x, y) { let res = "";{} return res; }\n')
      end)

      test("fn Name() -> res:text", function()
        parser:parse("fn sum(x?, y?) -> res:num")
        assert(trans:next()):eq('function sum(x, y) { let res = 0;{} return res; }\n')
      end)
    end)

    suite("standalone", function()
      suite("@alias", function()
        test("pub", function()
          parser:parse('@alias("f2")\npub fn f1(x) = x')
          assert(trans:next()):eq('function f1(x) { /* istanbul ignore next */ dogma.expect("x", x);{return x;} }export {f1}; const f2 = f1;export {f2};\n')
        end)

        test("not pub", function()
          parser:parse('@alias("f2")\nfn f1(x) = x')
          assert(trans:next()):eq('function f1(x) { /* istanbul ignore next */ dogma.expect("x", x);{return x;} } const f2 = f1;\n')
        end)
      end)

      test("@catch", function()
        parser:parse("@catch\nfn myfn()\n  return x + y")
        assert(trans:next()):eq("function myfn() { try { {return (x+y);} } catch { return null; } }\n")
      end)

      test("@iter fn Name() Body", function()
        parser:parse("@iter\nfn myfn()\n yield 1+2")
        assert(trans:next()):eq("function* myfn() { {yield (1+2);} }\n")
      end):tags("yield")

      test("@todo fn Name()", function()
        parser:parse("@todo\nfn myfn()")
        assert(trans:next()):eq("function myfn() { todo(); }\n")
      end)

      test("fn Name()", function()
        parser:parse("fn myfn()")
        assert(trans:next()):eq("function myfn() { {} }\n")
      end)

      test("fn Name() = Exp", function()
        parser:parse("fn myfn() = 123")
        assert(trans:next()):eq("function myfn() { {return 123;} }\n")
      end)

      test("export fn Name()", function()
        parser:parse("export fn myfn()")
        assert(trans:next()):eq('function myfn() { {} }module.exports = exports = myfn;\n')
      end)

      test("pub fn Name()", function()
        parser:parse("pub fn myfn()")
        assert(trans:next()):eq('function myfn() { {} }export {myfn};\n')
      end)

      test("fn Name(params) Body", function()
        parser:parse("fn sum(x, y)\n return x+y")
        assert(trans:next()):eq('function sum(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{return (x+y);} }\n')
      end)

      test("async fn Name(params) Body", function()
        parser:parse("async fn sum(x, y)\n return x+y")
        assert(trans:next()):eq('async function sum(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{return (x+y);} }\n')
      end)

      test("async fn Name(params) -> var Body", function()
        parser:parse("async fn sum(x, y) -> r\n r = x+y")
        assert(trans:next()):eq('async function sum(x, y) { let r;/* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{(r=(x+y));} return r; }\n')
      end)
    end)

    suite("type", function()
      suite("iter", function()
        test("@iter fn Name . iter () Body", function()
          parser:parse("@iter\nfn MyType.iter()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype[Symbol.iterator] = function*() {yield (1+2);};\n")
        end)

        test("@iter @todo fn Name . iter () Body", function()
          parser:parse("@iter @todo\nfn MyType.iter()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype[Symbol.iterator] = function*() { todo(); };\n")
        end)

        test("@iter fn Name . Name () Body", function()
          parser:parse("@iter\nfn MyType.iter2()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype.iter2 = function*() {yield (1+2);};\n")
        end)

        test("@iter @todo fn Name . Name () Body", function()
          parser:parse("@iter @todo\nfn MyType.iter2()\n yield 1+2")
          assert(trans:next()):eq("MyType.prototype.iter2 = function*() { todo(); };\n")
        end)
      end):tags("yield")

      suite("property", function()
        test("@prop fn Name . Name () = Exp", function()
          parser:parse("@prop\nfn MyType.prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { {return (1+2);} }});\n')
        end)

        test("fn Name . Name = Exp", function()
          parser:parse("fn MyType.prop = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { {return (1+2);} }});\n')
        end)

        test("@prop @todo fn Name . Name () = Exp", function()
          parser:parse("@prop @todo\nfn MyType.prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { todo(); }});\n')
        end)

        test("@prop fn Name . Name () -> Var Body", function()
          parser:parse("@prop\nfn MyType.prop() -> resp\n  resp = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { let resp;{(resp=(1+2));} return resp; }});\n')
        end)

        test("@prop fn Name : Name () = Exp", function()
          parser:parse("@prop\nfn MyType:prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "_prop", {enum: false, get: function() { {return (1+2);} }});\n')
        end)

        test("@prop fn Name ! Name () = Exp", function()
          parser:parse("@prop\nfn MyType!prop() = 1+2")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_prop", {enum: false, get: function() { {return (1+2);} }});\n')
        end)

        test("@abstract @prop fn Name.Name()", function()
          parser:parse("@abstract @prop\nfn MyType.prop()")
          assert(trans:next()):eq('Object.defineProperty(MyType.prototype, "prop", {enum: true, get: function() { /* istanbul ignore next */abstract(); }});\n')
        end)
      end):tags("prop")

      suite("static method", function()
        test("@static fn Name . Name() -> res Body", function()
          parser:parse("@static\nfn MyType.method() -> res\n  res = 1+2")
          assert(trans:next()):eq("MyType.method = function() { let res;{(res=(1+2));} return res; };\n")
        end)

        test('@static @alias("alias") fn Name . Name() -> res Body', function()
          parser:parse('@static @alias("func")\nfn MyType.method() -> res\n  res = 1+2')
          assert(trans:next()):eq("MyType.func = MyType.method = function() { let res;{(res=(1+2));} return res; };\n")
        end)

        test("@static async fn Name . Name() Body", function()
          parser:parse("@static\nasync fn MyType.method() = 1+2")
          assert(trans:next()):eq("MyType.method = async function() { {return (1+2);} };\n")
        end)

        test("@static @todo fn Name . Name() Body", function()
          parser:parse("@static @todo\nfn MyType.method() = 1+2")
          assert(trans:next()):eq("MyType.method = function() { todo(); };\n")
        end)
      end):tags("static")

      suite("abstract", function()
        test("pub", function()
          parser:parse("@abstract\nfn MyType.myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype.myfn = function() { abstract(); };\n')
        end)

        test("intl", function()
          parser:parse("@abstract\nfn MyType:myfn()")
          assert(trans:next()):eq('/* istanbul ignore next */ MyType.prototype._myfn = function() { abstract(); };\n')
        end)

        test("pvt", function()
          parser:parse("@abstract\nfn MyType!myfn()")
          assert(function() trans:next() end):raises("anonymous code (2, 1): abstract method can't be private (!).")
        end)
      end):tags("abstract")

      suite("instance", function()
        test("@alias", function()
          parser:parse('@alias("myfn2") fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.myfn2 = MyType.prototype.myfn = function() { {} };\n')
        end)

        test("async fn Name . Name ()", function()
          parser:parse("async fn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = async function() { {} };\n')
        end)

        test('@alias("alias") async fn Name . Name ()', function()
          parser:parse('@alias("mymeth") async fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.mymeth = MyType.prototype.myfn = async function() { {} };\n')
        end)

        test("fn Name . Name ()", function()
          parser:parse("fn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = function() { {} };\n')
        end)

        test('@alias("alias") fn Name . Name ()', function()
          parser:parse('@alias("mymeth") fn MyType.myfn()')
          assert(trans:next()):eq('MyType.prototype.mymeth = MyType.prototype.myfn = function() { {} };\n')
        end)

        test("@todo\\nfn Name . Name ()", function()
          parser:parse("@todo\nfn MyType.myfn()")
          assert(trans:next()):eq('MyType.prototype.myfn = function() { todo(); };\n')
        end)

        test("fn Name : Name ()", function()
          parser:parse("fn MyType:myfn()")
          assert(trans:next()):eq('MyType.prototype._myfn = function() { {} };\n')
        end)

        test("fn Name ! Name ()", function()
          parser:parse("fn MyType!myfn()")
          assert(trans:next()):eq('MyType.prototype._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_myfn = function() { {} };\n')
        end)

        test("fn Name . Name (. Name)", function()
          parser:parse("fn MyType.myfn(.x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name (. Name , . Name)", function()
          parser:parse("fn MyType.myfn(.x, .y)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x, y) { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);Object.defineProperty(this, "x", {value: x, enum: true, writable: true});Object.defineProperty(this, "y", {value: y, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name (const . Name)", function()
          parser:parse("fn MyType.myfn(const .x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
        end)

        test("fn Name . Name (const . Name)", function()
          parser:parse("fn MyType.myfn(const .x)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { /* istanbul ignore next */ dogma.expect("x", x);Object.defineProperty(this, "x", {value: x, enum: true, writable: false});{} };\n')
        end)

        test("fn Name . Name (.x Name ?)", function()
          parser:parse("fn MyType.myfn(.x?)")
          assert(trans:next()):eq('MyType.prototype.myfn = function(x) { Object.defineProperty(this, "x", {value: x, enum: true, writable: true});{} };\n')
        end)

        test("fn Name . Name() -> self\\n Body", function()
          parser:parse("fn MyType.method() -> self\n 1+2")
          assert(trans:next()):eq("MyType.prototype.method = function() { {(1+2);} return this; };\n")
        end)

        test("fn Name . Name() -> self\\n Body", function()
          parser:parse("fn MyType.method() -> self\n 1+2")
          assert(trans:next()):eq("MyType.prototype.method = function() { {(1+2);} return this; };\n")
        end)
      end):tags("instance")
    end)
  end):tags("fn")

  --------
  -- if --
  --------
  suite("if", function()
    test("#[!cov ignore] if Exp then...", function()
      parser:parse("#[!cov ignore] if true then x+y")
      assert(trans:next()):eq("/*istanbul ignore next*/if (true) {(x+y);}\n")
    end)

    test("#[!cov ignore else] if Exp then...", function()
      parser:parse("#[!cov ignore else] if true then x+y else y+x")
      assert(trans:next()):eq("/*istanbul ignore else*/if (true) {(x+y);} else {(y+x);}\n")
    end)

    test("#[!cov ignore] if const x = Exp then...", function()
      parser:parse("#[!cov ignore] if const x = true then x+y")
      assert(trans:next()):eq("{/*istanbul ignore next*/const x=true; /*istanbul ignore next*/if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore else] if const x = Exp then...", function()
      parser:parse("#[!cov ignore else] if const x = true then x+y")
      assert(trans:next()):eq("{const x=true; /*istanbul ignore else*/if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore] if x; y then...", function()
      parser:parse("#[!cov ignore] if x; y then x+y")
      assert(trans:next()):eq("/*istanbul ignore next*/x; /*istanbul ignore next*/if (y) {(x+y);}\n")
    end)

    test("#[!cov ignore else] if x; y then...", function()
      parser:parse("#[!cov ignore else] if x; y then x+y")
      assert(trans:next()):eq("x; /*istanbul ignore else*/if (y) {(x+y);}\n")
    end)

    test("if Exp then Exp", function()
      parser:parse("if true then x+y")
      assert(trans:next()):eq("if (true) {(x+y);}\n")
    end)

    test("if Exp; Exp then Exp", function()
      parser:parse("if x+=1; x==8 then x+y")
      assert(trans:next()):eq("(x+=1); if ((x==8)) {(x+y);}\n")
    end)

    test("if var Exp; Exp then Exp", function()
      parser:parse("if var x=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if Name := Name; Exp then Exp", function()
      parser:parse("if x:=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if Name := Name then Exp", function()
      parser:parse("if x:=1; x then x+y")
      assert(trans:next()):eq("{let x=1; if (x) {(x+y);}}\n")
    end)

    test("if [Name, Name] = Exp; Exp then Exp", function()
      parser:parse("if [x, y] = arr; x then x+y")
      assert(trans:next()):eq("([x, y]=arr); if (x) {(x+y);}\n")
    end)

    test("if [Name, Name] := Exp; Exp then Exp", function()
      parser:parse("if [x, y] := arr; x then x+y")
      assert(trans:next()):eq("{let [x, y] = arr; if (x) {(x+y);}}\n")
    end)

    test("#[!cov ignore next] if [Name, Name] := Exp; Exp then Exp", function()
      parser:parse("#[!cov ignore next] if [x, y] := arr; x then x+y")
      assert(trans:next()):eq("{/*istanbul ignore next*/let [x, y] = arr; /*istanbul ignore next*/if (x) {(x+y);}}\n")
    end)

    test("if [Name, Name] ::= Exp; Exp then Exp", function()
      parser:parse("if [x, y] ::= arr; x then x+y")
      assert(trans:next()):eq("{const [x, y] = arr; if (x) {(x+y);}}\n")
    end)

    test("if Name ::= Exp; Exp then Exp", function()
      parser:parse("if x::=1; x then x+y")
      assert(trans:next()):eq("{const x=1; if (x) {(x+y);}}\n")
    end)

    test("if const Exp; Exp then Exp", function()
      parser:parse("if const x=1; x then x+y")
      assert(trans:next()):eq("{const x=1; if (x) {(x+y);}}\n")
    end)

    test("if Exp then Sent", function()
      parser:parse("if true then return x+y")
      assert(trans:next()):eq("if (true) {return (x+y);}\n")
    end)

    test("if Exp then Exp else Exp", function()
      parser:parse("if true then x+y else a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else {(a+b);}\n")
    end)

    test("if Exp then Exp else Sent", function()
      parser:parse("if true then x+y else return a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else {return (a+b);}\n")
    end)

    test("if with else if", function()
      parser:parse("if true then\n x+y\nelse if false then\n a+b")
      assert(trans:next()):eq("if (true) {(x+y);} else if (false) {(a+b);}\n")
    end)

    test("if with else if and else", function()
      parser:parse("if x then\n x+y\nelse if y then\n a+b\nelse if z then\n 1+2\nelse\n 3+4")
      assert(trans:next()):eq("if (x) {(x+y);} else if (y) {(a+b);} else if (z) {(1+2);} else {(3+4);}\n")
    end)
  end):tags("if")

  ---------
  -- pub --
  ---------
  suite("pub", function()
    test("pub Item", function()
      parser:parse("pub Item")
      assert(trans:next()):eq("export {Item};\n")
    end)

    test('pub "./Item"', function()
      parser:parse('pub "./Item"')
      assert(trans:next()):eq('const Item = dogma.use(require("./Item"));export {Item};\n')
    end)

    test('pub Item, "./Item"', function()
      parser:parse('pub Item1, "./Item2"')
      assert(trans:next()):eq('export {Item1};const Item2 = dogma.use(require("./Item2"));export {Item2};\n')
    end)

    test("pub use(...)", function()
      parser:parse('pub use(\nItem1 as Item11\nItem22 = "./Item2"\n../Item3\n)')
      assert(trans:next()):eq(
        'const Item11 = dogma.use(require("./Item1"));export {Item11};' ..
        'const Item22 = dogma.use(require("./Item2"));export {Item22};' ..
        'const Item3 = dogma.use(require("../Item3"));export {Item3};' ..
        '\n'
      )
    end)
  end):tags("pub")

  ------------
  -- export --
  ------------
  suite("export", function()
    test("export Exp", function()
      parser:parse("export 1+2")
      assert(trans:next()):eq("module.exports = exports = (1+2);\n")
    end)
  end):tags("export")

  ----------
  -- with --
  ----------
  suite("with", function()
    test("simple cases", function()
      parser:parse("with 1+2 do\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2); switch (_) { case 1: {(x+1);};/*istanbul ignore next*/break; case 2: {(y+1);};/*istanbul ignore next*/break; default: {(z+1);}}}\n")
    end)

    test("composite cases", function()
      parser:parse("with 1+2 do\n  if 1, 2 then x+1\n  if 3 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2); switch (_) { case 1: case 2: {(x+1);};/*istanbul ignore next*/break; case 3: {(y+1);};/*istanbul ignore next*/break; default: {(z+1);}}}\n")
    end)

    test("with coverage directive: else", function()
      parser:parse("#!coverage ignore else\nwith 1+2 do\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2); switch (_) { case 1: {(x+1);};/*istanbul ignore next*/break; case 2: {(y+1);};/*istanbul ignore next*/break; /*istanbul ignore next*/ default: {(z+1);}}}\n")
    end)

    test("with coverage directive: next", function()
      parser:parse("#!coverage ignore next\nwith 1+2 do\n  if 1 then x+1\n  if 2 then y+1\n  else z+1")
      assert(trans:next()):eq("{const _ = (1+2);/*istanbul ignore next*/ switch (_) { case 1: {(x+1);};/*istanbul ignore next*/break; case 2: {(y+1);};/*istanbul ignore next*/break; default: {(z+1);}}}\n")
    end)

    test("with var := value do", function()
      parser:parse("with x := 1+2 do")
      assert(trans:next()):eq("{let x = (1+2); switch (x) {}}\n")
    end)

    test("with var ::= value do", function()
      parser:parse("with x ::= 1+2 do")
      assert(trans:next()):eq("{const x = (1+2); switch (x) {}}\n")
    end)

    test("with - with catch", function()
      parser:parse("with x ::= 1+2 do\n  if 1 then print(123)\ncatch e\n  print(e)")
      assert(trans:next()):eq("try {const x = (1+2); switch (x) { case 1: {print(123);};/*istanbul ignore next*/break;}} catch(e) {print(e);}\n")
    end)

    test("with - with finally", function()
      parser:parse("with x ::= 1+2 do\n  if 1 then print(123)\nfinally\n  print(e)")
      assert(trans:next()):eq("try {const x = (1+2); switch (x) { case 1: {print(123);};/*istanbul ignore next*/break;}} finally {print(e);}\n")
    end)
  end):tags("with")
end):tags("stmt")
